/*********************************************************************
 * Desktop Station one R3 Sketch for Mr Push R3
 *
 * Copyright (C) 2017 Yaasan
 *
 */

//#define DEBUG

// include the library code:
#include <EEPROM.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Wire.h"
#include "Functions.h"
#include "iSdio.h"
#include "Sd2CardExt.h"


#define MAX_S88DECODER 1
#define RELPYERROR_300 "300 Command error"
#define RELPYERROR_301 "301 Syntax error"
#define RELPYERROR_302 "302 receive timeout"
#define RELPYERROR_303 "303 Unknown error"
#define RELPYERROR_NONE ""

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_NOSD 3

#define PIN_RUNLED	2

#define SHAREMEM_SIZE 64

#define THRESHOLD_EDC 276  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_EDC_DSA 375  // 1024d * (1.2 / 8) * 9V / 5V
#define THRESHOLD_OV  553  // 1024d * (1.2 / 8) * 20V / 5V
#define THRESHOLD_OV_DSA  999  // 1024d * (1.2 / 8) * 20V / 5V

const int chipSelectPin = 4;
Sd2CardExt card;
uint8_t buffer_sharedMem[SHAREMEM_SIZE];
byte gFlag_FlashAir;

/* Decralation classes */
DSCoreLib DSCore;

/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Error Threshold */
word gThreshold_OV = 0;
word gThreshold_LV = 0;


/* Parser */
String request;
String function;
word arguments[4];
word numOfArguments;
word gLocAddresses[4] = {0,0,0,0};
word numOfLocAddr = 0;
boolean result;

//Task Schedule
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval
unsigned long gPreviousL9 = 0; // 2000ms interval



void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);
boolean SDIO_ReadSharedMem();
boolean SDIO_WriteSharedMem_Power(uint8_t aPower);
void SDIO_ClearSharedMem();


/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	
	
	DSCore.Init();

	//Pin mode
	pinMode(A1, INPUT);//Edc
	pinMode(PIN_RUNLED, OUTPUT);
	
	pinMode(A0, INPUT);
	digitalWrite(A0, HIGH);
	
	if( digitalRead(A0) == HIGH)
	{
		//DSair
		gThreshold_OV = THRESHOLD_OV_DSA;
		gThreshold_LV = THRESHOLD_EDC_DSA;
		
	}
	else
	{
		//DSshield with FlashAir
		gThreshold_OV = THRESHOLD_OV;
		gThreshold_LV = THRESHOLD_EDC;
	}
	
	
	
	/* Debug message for PC */
	Serial.begin(115200);
	Serial.println("--------------------------------------");
	Serial.println("Desktop Station FA R1b 2018/02/10     ");
	Serial.println("--------------------------------------");
	
	
	// Initialize SD card.
	Serial.print("Initializing SD card...");  
	
	if (card.init(SPI_FULL_SPEED, chipSelectPin)) 
	{
		Serial.println("OK");
		gFlag_FlashAir = 1;
		
	} else {
		Serial.println("NG");
		gFlag_FlashAir = 0;
		gLED_State = LEDSTATE_NOSD;
	}
	
	
	/* 線路に電源遮断 */
	DS_Power(0);
	
	cli();
	
	
	//タイマ0を強制的に戻すr
	TCCR0B = (TCCR0B & 0b11111000) | 0x03;
	//Timer1 init
	TCCR1A = 0;
	TCCR1B = 0;
	TCNT1  = 0;

	OCR1A = 22;            // 1Hz = 62500d, 100Hz = 625d
	TCCR1B |= (1 << WGM12);   // CTC mode
	TCCR1B |= (1 << CS12);    // 256 prescaler 
	TIMSK1 |= (1 << OCIE1A);  // enable timer compare interrupt

	/* LED turn on */
	changeLED();
	
	
	/* Wifi起動待ち */
	
	/*
	while()
	{
		digitalWrite(chipSelectPin, LOW); 
		uint8_t aResult = card.readExtMemory(1, 1, 0x1000, 32, buffer_sharedMem);
		digitalWrite(chipSelectPin, HIGH);
		
		digitalWrite(PIN_RUNLED, 0);
		delay(200);
		digitalWrite(PIN_RUNLED, 1);
		delay(100);
	}
	*/
	
	Serial.println("100 Ready");
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	
	
	sei();    //ENABLE INTERRUPTION

	
}


boolean parse(String inRequest) 
{
	int aPosSlash = -1;
	String request = inRequest;

	char aTemp[SHAREMEM_SIZE];
	request.toCharArray(aTemp, SHAREMEM_SIZE);

	int lpar = request.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(request.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = request.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;
	
	while (comma != -1) 
	{
		String tmp = request.substring(offset, comma);
		tmp.trim();
		
		/* スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);
		
		if( aPosSlash > 0)
		{
			tmp = tmp + "/";
			
			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;
			
			//スラッシュはアドレス部だけに許可
			while( aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();
				
				if( numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}
				
				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}
			
			offset = comma + 1;
			comma = request.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = request.indexOf(',', offset);
		}
	}

	int rpar = request.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = request.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	byte aValueHigh;
	boolean aResult;
	char aTempStr[16];
	word aLocProtocol;
	byte aDir = 0;
	long aSpeed = 0;
	
	if ((function == "setLocoDirection") || ((function == "DI")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				aResult =  DSCore.SetLocoDirection(gLocAddresses[i], aDir);
			}
			
			aResult = true;
		}
		else
		{
			//非重連
			aResult =  DSCore.SetLocoDirection(arguments[0], aDir);
		}
		
		return aResult;
	}
	else if ((function == "setLocoFunction") || ((function == "FN")))
	{
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "TO")))
	{
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "PW")))
	{
		
		if( arguments[0] == 1)
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
		}
		else
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
		}
		
		
		/* LED状態の変更 */
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "SP")))
	{
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				aResult =  DSCore.SetLocoSpeedEx(gLocAddresses[i], arguments[1], arguments[2]);
			}
			aResult = true;
		}
		else
		{
			//非重連
			aResult =  DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
		}
		
		return aResult;
	}
	else if ((function == "reset") || ((function == "RS")))
	{
		#ifdef DEBUG
		Serial.println("100 Ready");
		#endif
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "PG")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		#ifdef DEBUG
		Serial.println("@DSG,00,03,00,02,00,00,00,00");
		#endif
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "GV")))
	{
		
		return true;
	}
	else if (function == "getS88")
	{
		
		return true;
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "SV")))
	{
		
		return DSCore.WriteConfig_Dir( arguments[1], arguments[2]);
	}
	else
	{
		return false;
	}
}

/***********************************************************************
* 
* Timer Interval
* 
************************************************************************/

ISR(TIMER1_COMPA_vect)
{
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
}

void loop()
{
	word aEdc;
	uint8_t aErrNo = 0;
	String aRecvCmd = "";
	char *aReplyText = RELPYERROR_NONE;
	
	if( (millis() - gPreviousL8) >= 200)
	{
		//Reset task
		gPreviousL8 = millis();
		
		//LED Control
		taskLED();
		
		/* Sensor check */
		aEdc = analogRead(A1);
		
		/* OV check */
		if( aEdc > gThreshold_OV )
		{
			aErrNo = 2;
		}
		/* LV check */
		if( aEdc < gThreshold_LV )
		{
			aErrNo = 1;
			
		}
		/* エラー処理 */
		if( aErrNo > 0)
		{
			if( DSCore.IsPower() == 1)
			{
				PowerOffByErr(aErrNo);
			}
			
		}
	}
	
	if( SDIO_ReadSharedMem() == true)
	{
		//データをシリアルに流し込む
		
		for( int i = 0; i < SHAREMEM_SIZE; i++)
		{
			if( buffer_sharedMem[i] != ')' )
			{
				char aWord = (char)(buffer_sharedMem[i]);
				
				aRecvCmd = aRecvCmd + aWord;
			}
			else
			{
				aRecvCmd = aRecvCmd + ")";
				break;
			}
		}
		SDIO_ClearSharedMem();
		
		
		if (parse(aRecvCmd)) 
		{
			#ifdef DEBUG
			Serial.println(aRecvCmd);
			#endif
			if (dispatch()) 
			{
				aReplyText = "200 Ok";
			}
			else 
			{
				aReplyText = RELPYERROR_300;
			}
  			
		} 
		else 
		{
			aReplyText = RELPYERROR_301;
		}
		
		#ifdef DEBUG
		Serial.println(aReplyText);
		#endif
	}
	else
	{
		/* Nothing to do */
	}
}

void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	delay(1000);
	ReplyPowerPacket(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
}

void ReplyPowerPacket(byte inPower)
{
	#ifdef DEBUG
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower);
	Serial.println(",");
	#endif
}


void changeLED(void)
{
	
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, 0);
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, 1);
		break;
		
	case LEDSTATE_ERR:
		break;
		
	case LEDSTATE_NOSD:
		break;
	}
}

void taskLED(void)
{
	
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		digitalWrite(PIN_RUNLED, gLED_Toggle);
		gLED_Toggle = (~gLED_Toggle) & 1;
		break;
		
	case LEDSTATE_NOSD:
		
		if( gLED_Counter <= 0)
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else if(gLED_Counter == 2)
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		
		break;
	}
}



void SDIO_ClearSharedMem()
{
	
	//Clear Buffer
	memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	digitalWrite(chipSelectPin, LOW); 
	
	card.writeExtMemory(1, 1, 0x1000, SHAREMEM_SIZE, buffer_sharedMem);
	
	digitalWrite(chipSelectPin, HIGH); 
	
}

boolean SDIO_ReadSharedMem()
{
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	
	//Clear Buffer
	memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	digitalWrite(chipSelectPin, LOW); 
	
	uint8_t aResult = card.readExtMemory(1, 1, 0x1000, SHAREMEM_SIZE, buffer_sharedMem);
	
	digitalWrite(chipSelectPin, HIGH); 
	
	
	//Serial.print("->");
	//Serial.printDEC(buffer_sharedMem[0]);
	//Serial.println("");
	
	
	
	if (!aResult)
	{
		return false;
	}
	else
	{
		
		if( buffer_sharedMem[0] == 0x00)
		{
			return false;
		}
		else
		{
			return true;
		}
	}

}

boolean SDIO_WriteSharedMem_Power(uint8_t aPower)
{
	
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	//Clear Buffer
	memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	buffer_sharedMem[0] = aPower;
	
	digitalWrite(chipSelectPin, LOW); 
	
	uint8_t aResult = card.writeExtMemory(1, 1, 0x1080, 1, buffer_sharedMem);
	
	digitalWrite(chipSelectPin, HIGH); 
	
	if (!aResult)
	{
		return false;
	}
	else
	{
		return true;
	}

}
